@@ -1,34 +1,30 @@
 """
-:mod:`ngram` -- Provides a set that supports lookup by string similarity
-========================================================================
+:mod:`ngram` -- A set class that supports lookup by N-gram string similarity
+============================================================================
+"""
 
-.. moduleauthor:: Graham Poulter (version 3.0+)
-.. moduleauthor:: Michel Albert (version 2.0.0b2)
-"""
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU Lesser General Public License as published
+# by the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>
 
 from __future__ import division
 
-__license__ = """
-This library is free software; you can redistribute it and/or
-modify it under the terms of the GNU Lesser General Public
-License as published by the Free Software Foundation; either
-version 2.1 of the License, or (at your option) any later version.
-
-This library is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-Lesser General Public License for more details.
-
-See LICENSE file or http://www.gnu.org/licenses/lgpl-2.1.html
-"""
 
 class NGram(set):
-    """A set that supports lookup by NGram string similarity.
+    """A set that supports searching for members by N-gram string similarity.
 
-    Accepts `unicode` string or an encoded `str` of bytes. With encoded `str` the
-    splitting is on byte boundaries, which will be incorrect if the encoding uses
-    multiple bytes per character.  You must provide NGram with unicode strings if
-    the encoding would have multi-byte characters.
+    In Python 2, items should be `unicode` string or a plain ASCII `str`
+    (bytestring) - do not use UTF-8 or other multi-byte encodings, because
+    multi-byte characters will be split up.
 
     :type threshold: float in 0.0 ... 1.0
 
@@ -36,7 +32,8 @@ class NGram(set):
 
     :type warp: float in 1.0 ... 3.0
 
-    :param warp: use warp greater than 1.0 to increase the similarity of shorter string pairs.
+    :param warp: use warp greater than 1.0 to increase the similarity of \
+    shorter string pairs.
 
     :type items: [item, ...]
 
@@ -52,23 +49,26 @@ class NGram(set):
 
     :type pad_char: str or unicode
 
-    :param pad_char: character to use for padding.  Default is '$', but consider using the\
-    non-breaking space character, ``u'\\xa0'`` (``u"\\u00A0"``).
+    :param pad_char: character to use for padding.  Default is '$', but \
+    consider using the\ non-breaking space character, ``u'\\xa0'`` \
+    (``u"\\u00A0"``).
 
     :type key: function(item) -> str/unicode
 
-    :param key: Function to convert items into string, default is no conversion.
+    :param key: Function to convert items into string, default is no \
+    conversion.  Recommended to use `str` or `unicode` for non-string items. \
+    Using anonymous function prevents NGram class from being pickled.
 
     Instance variables:
 
-    :ivar _grams: For each n-gram, the items containing it and the number of times\
-    the n-gram occurs in the item as ``{str:{item:int, ...}, ...}``.
+    :ivar _grams: For each n-gram, the items containing it and the number of \
+    times\ the n-gram occurs in the item as ``{str:{item:int, ...}, ...}``.
 
-    :ivar length: maps items to length of the padded string representations as
-    ``{item:int, ...}``.
+    :ivar length: maps items to length of the padded string representations \
+    as ``{item:int, ...}``.
     """
 
-    def __init__(self, items=[], threshold=0.0, warp=1.0, key=None,
+    def __init__(self, items=None, threshold=0.0, warp=1.0, key=None,
                     N=3, pad_len=None, pad_char='$', **kwargs):
         super(NGram, self).__init__()
         if not (0 <= threshold <= 1):
@@ -80,10 +80,10 @@ class NGram(set):
         if not N >= 1:
             raise ValueError("N out of range (should be N >= 1): " + repr(N))
         if pad_len is None:
-            pad_len = N-1
+            pad_len = N - 1
         if not (0 <= pad_len < N):
             raise ValueError("pad_len out of range: " + repr(pad_len))
-        if not len(pad_char)==1:
+        if not len(pad_char) == 1:
             raise ValueError(
                 "pad_char is not single character: " + repr(pad_char))
         if key is not None and not callable(key):
@@ -93,7 +93,7 @@ class NGram(set):
         self.N = N
         self._pad_len = pad_len
         self._pad_char = pad_char
-        self._padding = pad_char * pad_len # derive a padding string
+        self._padding = pad_char * pad_len  # derive a padding string
         # compatibility shim for 3.1 iconv parameter
         if 'iconv' in kwargs:
             self._key = kwargs.pop('iconv')
@@ -104,13 +104,15 @@ class NGram(set):
         self._key = key
         self._grams = {}
         self.length = {}
-        self.update(items)
+        if items:
+            self.update(items)
 
     def __reduce__(self):
-        """Return state information for pickling, no references to this instance.
-        The key function must be None, a builtin function, or a named
-        module-level function.
+        """Return state information for pickling, no references to this
+        instance.  The key function must be None, a builtin function, or
+        a named module-level function.
 
+        >>> from ngram import NGram
         >>> n = NGram([0xDEADBEEF, 0xBEEF], key=hex)
         >>> import pickle
         >>> p = pickle.dumps(n)
@@ -121,10 +123,13 @@ class NGram(set):
         return NGram, (list(self), self.threshold, self.warp, self._key,
                        self.N, self._pad_len, self._pad_char)
 
-    def copy(self):
-        """Return a shallow copy of the NGram object.  That is, instantiate
-        a new NGram from references to items stored in this one.
+    def copy(self, items=None):
+        """Return a new NGram object with the same settings, and
+        referencing the same items.  Copy is shallow in that
+        each item is not recursively copied.   Optionally specify
+        alternate items to populate the copy.
 
+        >>> from ngram import NGram
         >>> from copy import deepcopy
         >>> n = NGram(['eggs', 'spam'])
         >>> m = n.copy()
@@ -133,13 +138,18 @@ class NGram(set):
         ['eggs', 'spam']
         >>> list(m)
         ['eggs', 'ham', 'spam']
+        >>> p = n.copy(['foo', 'bar'])
+        >>> list(p)
+        ['foo', 'bar']
         """
-        return NGram(self, self.threshold, self.warp, self._key,
+        return NGram(items if items is not None else self,
+                     self.threshold, self.warp, self._key,
                      self.N, self._pad_len, self._pad_char)
 
     def key(self, item):
         """Get the key string for the item.
 
+        >>> from ngram import NGram
         >>> n = NGram(key=lambda x:x[1])
         >>> n.key((3,"ham"))
         'ham'
@@ -149,6 +159,7 @@ class NGram(set):
     def pad(self, string):
         """Pad a string in preparation for splitting into ngrams.
 
+        >>> from ngram import NGram
         >>> n = NGram()
         >>> n.pad('ham')
         '$$ham$$'
@@ -158,12 +169,13 @@ class NGram(set):
     def _split(self, string):
         """Iterates over the ngrams of a string (no padding).
 
+        >>> from ngram import NGram
         >>> n = NGram()
         >>> list(n._split("hamegg"))
         ['ham', 'ame', 'meg', 'egg']
         """
         for i in range(len(string) - self.N + 1):
-            yield string[i:i+self.N]
+            yield string[i:i + self.N]
 
     # compatibility with 3.1
     ngrams = _split
@@ -171,6 +183,7 @@ class NGram(set):
     def split(self, string):
         """Pads a string and iterates over its ngrams.
 
+        >>> from ngram import NGram
         >>> n = NGram()
         >>> list(n.split("ham"))
         ['$$h', '$ha', 'ham', 'am$', 'm$$']
@@ -183,6 +196,7 @@ class NGram(set):
     def splititem(self, item):
         """Pads the string key of an item and iterates over its ngrams.
 
+        >>> from ngram import NGram
         >>> n = NGram(key=lambda x:x[1])
         >>> item = (3,"ham")
         >>> list(n.splititem(item))
@@ -191,8 +205,9 @@ class NGram(set):
         return self.split(self.key(item))
 
     def add(self, item):
-        """Add an item to the N-gram index (only if it has not already been added).
+        """Add an item to the N-gram index (if it has not already been added).
 
+        >>> from ngram import NGram
         >>> n = NGram()
         >>> n.add("ham")
         >>> list(n)
@@ -214,8 +229,9 @@ class NGram(set):
                 self._grams[ngram][item] += 1
 
     def remove(self, item):
-        """Remove an item from the index. Inverts the add operation.
+        """Remove an item from the set. Inverts the add operation.
 
+        >>> from ngram import NGram
         >>> n = NGram(['spam', 'eggs'])
         >>> n.remove('spam')
         >>> list(n)
@@ -227,12 +243,28 @@ class NGram(set):
             for ngram in self.splititem(item):
                 del self._grams[ngram][item]
 
+    def pop(self):
+        """Remove and return an arbitrary set element.
+        Raises KeyError if the set is empty.
+
+        >>> from ngram import NGram
+        >>> n = NGram(['spam', 'eggs'])
+        >>> n.pop()
+        'eggs'
+        """
+        item = super(NGram, self).pop()
+        del self.length[item]
+        for ngram in self.splititem(item):
+            del self._grams[ngram][item]
+        return item
+
     def items_sharing_ngrams(self, query):
         """Retrieve the subset of items that share n-grams the query string.
 
         :param query: look up items that share N-grams with this string.
-        :return: dictionary from matched string to the number of shared N-grams.
+        :return: mapping from matched string to the number of shared N-grams.
 
+        >>> from ngram import NGram
         >>> n = NGram(["ham","spam","eggs"])
         >>> n.items_sharing_ngrams("mam")
         {'ham': 2, 'spam': 2}
@@ -246,7 +278,7 @@ class NGram(set):
             try:
                 for match, count in self._grams[ngram].items():
                     remaining.setdefault(ngram, {}).setdefault(match, count)
-                    # match up to as many occurrences of ngram as exist in the matched string
+                    # match as many occurrences as exist in matched string
                     if remaining[ngram][match] > 0:
                         remaining[ngram][match] -= 1
                         shared.setdefault(match, 0)
@@ -262,18 +294,19 @@ class NGram(set):
         :return: list of pairs of (item, similarity) by decreasing similarity.
 
         >>> from ngram import NGram
-        >>> n = NGram([(0, "SPAM"), (1, "SPAN"), (2, "EG")], key=lambda x:x[1])
-        >>> n.searchitem((2, "SPA"))
+        >>> n = NGram([(0, "SPAM"), (1, "SPAN"), (2, "EG"),
+        ... (3, "SPANN")], key=lambda x:x[1])
+        >>> n.searchitem((2, "SPA"), 0.35)
         [((0, 'SPAM'), 0.375), ((1, 'SPAN'), 0.375)]
         """
-        return self.search(self.key(item))
+        return self.search(self.key(item), threshold)
 
     def search(self, query, threshold=None):
         """Search the index for items whose key exceeds threshold
         similarity to the query string.
 
-        :param query: returned items will have at least `threshold` similarity to
-        the query string.
+        :param query: returned items will have at least `threshold` \
+        similarity to the query string.
 
         :return: list of pairs of (item, similarity) by decreasing similarity.
 
@@ -296,7 +329,7 @@ class NGram(set):
             if similarity >= threshold:
                 results.append((match, similarity))
         # Sort results by decreasing similarity
-        results.sort(key=lambda x:x[1], reverse=True)
+        results.sort(key=lambda x: x[1], reverse=True)
         return results
 
     def finditem(self, item, threshold=None):
@@ -304,11 +337,13 @@ class NGram(set):
         nothing exceeds the threshold.
 
         >>> from ngram import NGram
-        >>> n = NGram([(0, "Spam"), (1, "Ham"), (2, "Eggsy")], key=lambda x:x[1].lower())
+        >>> n = NGram([(0, "Spam"), (1, "Ham"), (2, "Eggsy"), (3, "Egggsy")],
+        ...     key=lambda x:x[1].lower())
         >>> n.finditem((3, 'Hom'))
         (1, 'Ham')
         >>> n.finditem((4, "Oggsy"))
         (2, 'Eggsy')
+        >>> n.finditem((4, "Oggsy"), 0.8)
         """
         results = self.searchitem(item, threshold)
         if results:
@@ -325,6 +360,7 @@ class NGram(set):
         'Ham'
         >>> n.find("Spom")
         'Spam'
+        >>> n.find("Spom", 0.8)
         """
         results = self.search(query, threshold)
         if results:
@@ -336,8 +372,8 @@ class NGram(set):
     def ngram_similarity(samegrams, allgrams, warp=1.0):
         """Similarity for two sets of n-grams.
 
-        :note: ``similarity = (a**e - d**e)/a**e`` where `a` is "all n-grams",
-        `d` is "different n-grams" and `e` is the warp.
+        :note: ``similarity = (a**e - d**e)/a**e`` where `a` is \
+        "all n-grams", `d` is "different n-grams" and `e` is the warp.
 
         :param samegrams: number of n-grams shared by the two strings.
 
@@ -356,11 +392,12 @@ class NGram(set):
         >>> NGram.ngram_similarity(3, 4)
         0.75
         """
-        if abs(warp-1.0) < 1e-9:
+        if abs(warp - 1.0) < 1e-9:
             similarity = float(samegrams) / allgrams
         else:
             diffgrams = float(allgrams - samegrams)
-            similarity = (allgrams**warp - diffgrams**warp) / (allgrams**warp)
+            similarity = ((allgrams ** warp - diffgrams ** warp)
+                    / (allgrams ** warp))
         return similarity
 
     @staticmethod
@@ -396,6 +433,7 @@ class NGram(set):
     def update(self, items):
         """Update the set with new items.
 
+        >>> from ngram import NGram
         >>> n = NGram(["spam"])
         >>> n.update(["eggs"])
         >>> list(n)
@@ -405,8 +443,11 @@ class NGram(set):
             self.add(item)
 
     def discard(self, item):
-        """If `item` is a member of the set, remove it.
+        """Remove an element from a set if it is a member.
 
+        If the element is not a member, do nothing.
+
+        >>> from ngram import NGram
         >>> n = NGram(['spam', 'eggs'])
         >>> n.discard('spam')
         >>> n.discard('ham')
@@ -416,38 +457,100 @@ class NGram(set):
         if item in self:
             self.remove(item)
 
+    def clear(self):
+        """Remove all elements from this set.
+
+        >>> from ngram import NGram
+        >>> n = NGram(['spam', 'eggs'])
+        >>> list(n)
+        ['eggs', 'spam']
+        >>> n.clear()
+        >>> list(n)
+        []
+        """
+        super(NGram, self).clear()
+        self._grams = {}
+        self.length = {}
+
+    def union(self, *others):
+        """Return the union of two or more sets as a new set.
+
+        >>> from ngram import NGram
+        >>> a = NGram(['spam', 'eggs'])
+        >>> b = NGram(['spam', 'ham'])
+        >>> list(a.union(b))
+        ['eggs', 'ham', 'spam']
+        """
+        return self.copy(super(NGram, self).union(*others))
+
+    def difference(self, *others):
+        """Return the difference of two or more sets as a new set.
+
+        >>> from ngram import NGram
+        >>> a = NGram(['spam', 'eggs'])
+        >>> b = NGram(['spam', 'ham'])
+        >>> list(a.difference(b))
+        ['eggs']
+        """
+        return self.copy(super(NGram, self).difference(*others))
+
     def difference_update(self, other):
         """Remove from this set all elements from `other` set.
 
+        >>> from ngram import NGram
         >>> n = NGram(['spam', 'eggs'])
         >>> other = set(['spam'])
         >>> n.difference_update(other)
         >>> list(n)
         ['eggs']
         """
-        for x in other:
-            self.discard(x)
+        for item in other:
+            self.discard(item)
 
-    def intersection_update(self, other):
-        """Update the set with the intersection of itself and `other`.
+    def intersection(self, *others):
+        """Return the intersection of two or more sets as a new set.
 
+        >>> from ngram import NGram
+        >>> a = NGram(['spam', 'eggs'])
+        >>> b = NGram(['spam', 'ham'])
+        >>> list(a.intersection(b))
+        ['spam']
+        """
+        return self.copy(super(NGram, self).intersection(*others))
+
+    def intersection_update(self, *others):
+        """Update the set with the intersection of itself and other sets.
+
+        >>> from ngram import NGram
         >>> n = NGram(['spam', 'eggs'])
         >>> other = set(['spam', 'ham'])
         >>> n.intersection_update(other)
         >>> list(n)
         ['spam']
         """
-        self.difference_update([x for x in self if x not in other])
+        self.difference_update(super(NGram, self).difference(*others))
+
+    def symmetric_difference(self, other):
+        """Return the symmetric difference of two sets as a new set.
+
+        >>> from ngram import NGram
+        >>> a = NGram(['spam', 'eggs'])
+        >>> b = NGram(['spam', 'ham'])
+        >>> list(a.symmetric_difference(b))
+        ['eggs', 'ham']
+        """
+        return self.copy(super(NGram, self).symmetric_difference(other))
 
     def symmetric_difference_update(self, other):
         """Update the set with the symmetric difference of itself and `other`.
 
+        >>> from ngram import NGram
         >>> n = NGram(['spam', 'eggs'])
         >>> other = set(['spam', 'ham'])
         >>> n.symmetric_difference_update(other)
         >>> list(n)
         ['eggs', 'ham']
         """
-        intersection = self.intersection(other) # record intersection of sets
-        self.update(other) # add items present in other
-        self.difference_update(intersection) # remove items present in both
+        intersection = super(NGram, self).intersection(other)
+        self.update(other)  # add items present in other
+        self.difference_update(intersection)  # remove items present in both
